(define-package "git-commit" "3.3.0"
  "Edit Git commit messages."
  '((emacs "25.1")
    (dash "2.19.1")
    (transient "0.3.6")
    (with-editor "3.0.5"))
  :homepage "https://magit.vc"
  :keywords '("git" "tools" "vc"))
