(define-package "magit" "3.3.0"
  "A Git porcelain inside Emacs."
  '((emacs "25.1")
    (dash "2.19.1")
    (git-commit "3.3.0")
    (magit-section "3.3.0")
    (transient "0.3.6")
    (with-editor "3.0.5"))
  :homepage "https://magit.vc"
  :keywords '("git" "tools" "vc"))
